#!/bin/bash
# First, find out where we are
case $0 in
         /*)  SHELLFILE=$0 ;;
		./*)  SHELLFILE=${PWD}${0#.} ;;
        ../*) SHELLFILE=${PWD%/*}${0#..} ;;
          *)  SHELLFILE=$(type -P "$0") ; if [ "${SHELLFILE:0:1}" != "/" ]; then SHELLFILE="${PWD}/$SHELLFILE" ; fi ;;
esac

# readlink -m makes sure SHELLDIR is a simple absolute path
SHELLDIR=`readlink -m "${SHELLFILE%/*}"`

# So the RT-STPS root directory should be one directory above
RT_STPS_HOME="$SHELLDIR/.."

ARCH=`uname -m`
if [ ${ARCH} == 'x86_64' ]; then

# 64-bit linux
HDF="$RT_STPS_HOME/lib/hdf-java-2.7/linux64/hdf-java/lib"
HDF_LIB="linux"

elif [ $(ARCH) == 'ia64' ]; then

# 64-bit itanium
HDF="$RT_STPS_HOME/lib/hdf-java-2.7/linux64i/hdf-java/lib"
HDF_LIB="linux"

else 

# 32-bit linux
HDF="$RT_STPS_HOME/lib/hdf-java-2.7/linux32/hdf-java/lib"
HDF_LIB="linux"

fi

######################################################################################
# Functions:
######################################################################################
# Complain and exit function

function printUsageAndExit () {
	echo "Usage:"
	echo "   server.sh <server name> <RMI port> <data port> <control port>"
	echo "OPTIONAL arguments:"
	echo "   <server name> - the desired name of the new RT-STPS server instance to start, in double-quotes "
	echo "   <RMI port> - the desired port number of the new RT-STPS server's RMI registry. Must be provided for non-default <server name> "
	echo "   <data port> - the desired port number of the new RT-STPS server's data input port. Must be provided for non-default <server name> "
	echo "   <control port> - the desired port number of the new RT-STPS server's alternate interface. Must be provided for non-default <server name> "
	echo ""
	echo "Example usage:"
	echo "  ./server.sh"
	echo "  ./server.sh \"rtstps2\" 5000 5001 5002"
	echo ""
	echo "WARNING - The following values are reserved for the default RT-STPS server instance and may NOT be used"
	echo "for additional, non-default instances:"
	echo "server name: \"A\", port numbers: 1099, 4935, 5935"
	exit 1
}

function printUsage () {
	echo "WARNING - The following values are reserved for the default RT-STPS server instance and may NOT be used for additional, non-default instances:"
	echo "server name: \"A\", port numbers: 1099, 4935, 5935"
}

######################################################################################
# Script starts here:
######################################################################################
# Initialize the input parameters to default values
servername="A"
rmiport="1099"
dataport="4935"
controlport="5935"

# Check user arguments and perform appropriate actions
if [ "$#" -ge 4 ]; then

	# First check if an RT-STPS server with the same server name is already running
	sameServerRunning=`ps aux | grep "rt-stps" | grep -c "gov.nasa.gsfc.drl.rtstps.server.TcpServer $1"`
	if [ "$sameServerRunning" -gt 0 ]; then
		echo "WARNING - RT-STPS server instance \"$1\" is already running; please stop this server instance before continuing"
		echo "Command: $RT_STPS_HOME/bin/stop.sh \"$1\""
		exit 1
	fi

	# Assign new parameter values
	servername="$1"
	rmiport="$2"
	dataport="$3"
	controlport="$4"

	# Then, check if there are already instances that share the same server name, do not continue if there are
	if [ -d "$RT_STPS_HOME/jsw_$servername/" ]; then
		echo "WARNING - Cannot start new server instance; RT-STPS JSW server instance \"$servername\" already exists!"
		echo "JSW Server info location: $RT_STPS_HOME/jsw_$servername/"
		exit 1
	elif [ -f "$RT_STPS_HOME/info/server_info_$servername" ]; then
		echo "WARNING - Cannot start new server instance; RT-STPS server instance \"$servername\" already exists!"
		echo "Server info location: $RT_STPS_HOME/info/server_info_$servername"
		exit 1
	fi

	# Now check the parameters and see if they use any reserved values. Immediately complain and exit if any do.
	defvaluesfound=0
	if [ "$servername" == "A" ]; then
		defvaluesfound=1
	elif [ "$rmiport" -eq 1099 ] || [ "$rmiport" -eq 4935 ] || [ "$rmiport" -eq 5935 ]; then
		defvaluesfound=1
	elif [ "$dataport" -eq 1099 ] || [ "$dataport" -eq 4935 ] || [ "$dataport" -eq 5935 ]; then
		defvaluesfound=1
	elif [ "$controlport" -eq 1099 ] || [ "$controlport" -eq 4935 ] || [ "$controlport" -eq 5935 ]; then
		defvaluesfound=1
	fi

	if [ "$defvaluesfound" -gt 0 ]; then
		echo ""
		echo "WARNING - Cannot start new server instance; one or more of the provided arguments use values reserved for the default RT-STPS JSW server instance!"
		echo "WARNING - Arguments provided: $1 $2 $3 $4"
		printUsage
		exit 1
	fi

	# If parameters do not conflict with reserved values, record them into a server info file
	echo "servername=$servername" > "$RT_STPS_HOME/info/server_info_$servername"
	echo "rmiport=$rmiport" >> "$RT_STPS_HOME/info/server_info_$servername"
	echo "dataport=$dataport" >> "$RT_STPS_HOME/info/server_info_$servername"
	echo "controlport=$controlport" >> "$RT_STPS_HOME/info/server_info_$servername"
	echo "INFO - Starting RT-STPS server with the following parameters: server name = $servername, rmi port = $rmiport, data port = $dataport, control port = $controlport"
elif [ "$#" -eq 0 ]; then
	echo "INFO - Starting RT-STPS server with default parameters"
else
	echo "ERROR - When starting an RT-STPS server instance with non-default parameters, please make sure all optional parameters are provided"
	printUsageAndExit
fi

# Since there is no JSW to handle all the stdout messages and perform proper
# log rotation and deletion (to prevent build-up of log files), disable all stdout
# messages; only log stderr messages.
if [ "$servername" == "A" ]; then
	rtstpslog="$RT_STPS_HOME/logs/rt-stps-server.log"
else
	rtstpslog="$RT_STPS_HOME/logs/rt-stps-server_$servername.log"
fi

# This will have the value of RT-STPS home, but potential spaces are encoded for URL-compliance (for RMI purposes)
rmiURLDIR=`echo $RT_STPS_HOME | sed -e "s:\ :%20:g"`

java -server -Djava.library.path="$HDF/$HDF_LIB" -cp "$RT_STPS_HOME/lib/rt-stps.jar:$RT_STPS_HOME/lib/nsls.jar:$HDF/fits.jar:$HDF/jhdf.jar:$HDF/jhdf4obj.jar:$HDF/jhdf5.jar:$HDF/jhdf5obj.jar:$HDF/jhdfobj.jar:$HDF/jhdfview.jar:$HDF/junit.jar:$HDF/netcdf.jar:$HDF/ext/fitsobj.jar:$HDF/ext/nc2obj.jar" -Djava.rmi.server.codebase="file:$rmiURLDIR/lib/rt-stps.jar" -Dsecurity=0 -Dport=$dataport -DcontrolPort=$controlport -DrmiPort=$rmiport -DbufferSizeKb=16 -Dsetup="$RT_STPS_HOME/config" -Dlog.file="$rtstpslog" -Droot="$RT_STPS_HOME" gov.nasa.gsfc.drl.rtstps.server.TcpServer $servername > /dev/null 2>"$rtstpslog"

# If this was terminated unnaturally, blow away its corresponding server info
# (Because unlike in Windows, in Linux the server is not automatically put to background)
if [ "$?" -ne 0 ]; then
	rm -f "$RT_STPS_HOME/info/server_info_$servername" >/dev/null 2>&1
else
	echo "INFO - RT-STPS server successfully stopped"
fi
