/*
Copyright (c) 1999-2007, United States Government, as represented by
the Administrator for The National Aeronautics and Space Administration.
All rights reserved.
*/
package gov.nasa.gsfc.drl.rtstps.core.output.rdr;

import java.util.EnumSet;

import gov.nasa.gsfc.drl.rtstps.core.RtStpsException;
import gov.nasa.gsfc.drl.rtstps.core.ccsds.Packet;

/**
 * Build the RawApplicationPackets (or read a pre-existing one) for the HDF for ATMS.
 * This version uses the scan start bit to collect scan lines to make granules of 32 secs
 * worth approximately.
 * 
 * Current 
 *
 */
public class ATMSRawApplicationPackets3 extends RawApplicationPackets {
	// 5.7: Removed unnecessary stdout messages to reduce clutter in logs
    	private long baseTime = 0l;
    	private boolean fullGranule = false;
    	private long firstTime = -1l;
    	private long currentScanNumber = -1;
	private int scanCounter = 0;

	// 6.0: Removed old Packet Name enum sets no longer used

	private Stats stats;
	private int sciPacketCount = 0;
	private long lastCalPacketScanNumber = -999;
	private long lastEngPacketScanNumber = -999;
	private int lastEngPacketIndex;
	private int lastCalPacketIndex;
	private final int scansPerGranule = 12; // approx 32 secs worth
	//private final int sciPacketsPerScan = 104;

	private int scanCount = -1;  // FIXME there was a bug in VIIRS module and this should be zero and check below should >= I think...
	
	private static long currentgranule_startboundary=-1L;
	private static long currentgranule_endboundary=-1L;
	private static long granuleSize=31997000;
	//maximum number of VIIRS packets is the following:
	//48 scans * (479 + 17 DNB_MGS + 17 DNG_LGS)  packets =  24624 packets
	//for atms: 12 scans * ( 104 )
	private static long ATMSGRANULE_MAXPACKETS=1268L;
	
	// 6.0: Changing this method to use the appropriate packet name methods for comparison
	private boolean isATMSSciencePacket(Packet p) {
		// 6.0: Get the packet name object associated with the packet's APID
		PacketName atmsPacket = PacketName.fromAppId(p.getApplicationId());

		// 6.0: Return false if null
		if(atmsPacket == null) return false;

		// 6.0: Though ATMS-SCIENCE-RDR is composed of 4 APIDs, this implementation of RawApplicationPackets
		// only considers "SCI" as the "Science" packet. This is so granule boundaries are calculated
		// solely based on SCI packets, as including the other 3 APIDs in the equation results in
		// erroneous granule boundary calculations.
		return atmsPacket.getApidShortName().equals("SCI");
	}

	// 6.0: Removed old enum set check methods that are no longer used
	
	//private boolean scanCheck = false;

	/**
	 * Constructor for creating an nth instance of a ATMS raw application data packet area builder.
	 * @param satellite  the name of the spacecraft as a SpacecraftId
	 * @param setNum  the set number
	 * @param scansPerGranule the number of sensor scans per granule
	 */
	public ATMSRawApplicationPackets3(SpacecraftId satellite, int setNum, int scansPerGranule, PacketPool packetPool) {
		// 6.0: RDR Name is now a class and not an enum value
		super(satellite, RDRName.fromRDRNameStr("ATMS-SCIENCE-RDR"), setNum, packetPool);
		
		if (setNum < 0) {
			throw new IllegalArgumentException("Illegal Index [" + setNum + "]");
		}
	}
	

	public ATMSRawApplicationPackets3(Stats stats, SpacecraftId satellite, int setNum, int scansPerGranule, PacketPool packetPool) {
		// 6.0: RDR Name is now a class and not an enum value
		super(satellite, RDRName.fromRDRNameStr("ATMS-SCIENCE-RDR"), setNum, packetPool);
		
		if (setNum < 0) {
			throw new IllegalArgumentException("Illegal Index [" + setNum + "]");
		}
		this.stats = stats;
	}
	
	/**
	 * Constructor which attempts to read the RawApplicationPacket entry that pre-exists.
	 * The contents of the dataspace are read into a memory buffer... assuming it will fit.
	 * @param allRDRId  the rdrAll Groups id
	 * @param setNum the set number of raw entry
	 * @throws RtStpsException Wraps any HDF exception
	 */
	public ATMSRawApplicationPackets3(int allRDRId, int setNum) throws RtStpsException {
		super(allRDRId, setNum);
	}
	
	public ATMSRawApplicationPackets3(int readId, int setNum, boolean usedByGranuleOnly) throws RtStpsException {
		super(readId, setNum, true);
	}
	

	public boolean notFull(Packet p) throws RtStpsException 
	{
	    if(!isATMSSciencePacket(p))
			return true;
	    
	    if (CRISSciencePacket.checkForScanStart(p))
		scanCounter++;
	    
		// 6.0: Removed the rdr name check here as it is redundant (thanks to RDRCreateFile) and may cause race conditions
	    
	    boolean notFullStatus = true;
	    
	    //System.out.println("S/C APID:"+p.getApplicationId()+":"+p.getTimeStamp(8));
		
	    long ietTimeCurrentPacket=LeapDate.getMicrosSinceEpoch(p.getTimeStamp(8));
	    long timestamp = p.getTimeStamp(8);
			
	    if(this.firstTime < 0l) //Starting a granule
	    {
			if(this.baseTime == 0l)
			{
				//System.out.println("Base time was not set.");
				ATMSGranule.setBaseTime();
				this.baseTime = ATMSGranule.getBaseTime();
				//SpacecraftDiaryGranule.setBaseTime(ietTimeCurrentPacket);
				this.firstTime = timestamp;
				currentgranule_startboundary=ATMSGranule.getStartBoundary(ietTimeCurrentPacket);
				currentgranule_endboundary = ATMSGranule.getEndBoundary(ietTimeCurrentPacket);
				//System.out.println("Setting ATMS startBoundary and endBoundary"+ietTimeCurrentPacket+" "+currentgranule_startboundary+" "+currentgranule_endboundary);
				
			}
	    }
	    //If the ietTime is > startBoundary and <=endBoundary (or should it be >=startBoundary and < endBoundary)
	    //then granule is not full; else it is full
	    if(ietTimeCurrentPacket>=currentgranule_startboundary && ietTimeCurrentPacket<currentgranule_endboundary)
	    {
			notFullStatus = true;
			//System.out.println("Not full ATMS granule.");
	    }
	    else
	    {			
			notFullStatus = false;
			//System.out.println("Full ATMS granule.");
			//System.out.println("iettimecurrentpacket: " + ietTimeCurrentPacket);
			//System.out.println("currentgranule_startboundary: " + currentgranule_startboundary);
			//System.out.println("currentgranule_endboundary: " + currentgranule_endboundary);
			fullGranule = true;
			this.firstTime = -1l;
	    }
	   
	    return notFullStatus;
	    
	}
		
	@Override
	public void put(Packet p) throws RtStpsException 
	{
		
	    if (scanCounter < 0) { // throw away any initial packets if we've not seen the start of a scan...
		if (stats == null) {
			//System.out.println("Trimming ATMS [" + p.getApplicationId() + "]");
		} else {
			//stats.sci_trimmedPackets.value++;
		}
		return;
	    }
	    sciPacketCount++;
	
	    this.updateAppIdCounters(p.getApplicationId());
	
		// 6.0: Removed the rdr name check here as it is redundant (thanks to RDRCreateFile) and may cause race conditions
	    
	    // collect time info
	    if (p.hasSecondaryHeader()) {
			// secondary time stamp only valid in in first or standalone packets...
			if (getFirstTime() == 0L) {
				setFirstTime(p.getTimeStamp(8));
			}
		
			setLastTime(p.getTimeStamp(8));
	    }

	    Packet pcopy = CopyPacket.deep(p, packetPool);
	
	    getPacketList().add(pcopy);
		
	}
	
	
	/**
	 * Write the ATMSRawApplication structure to the given HDF file specified by the HDF input handle
	 * @param hdfFile input handle to a previous opened HDF file
	 * @return true if the write succeeds
	 * @throws RtStpsException Wraps HDF exceptions
	 */
	@Override
	public boolean write(int hdfFile) throws RtStpsException {
		boolean ret = false;
		float percentMissingData= ((ATMSGRANULE_MAXPACKETS-sciPacketCount)/ (float)ATMSGRANULE_MAXPACKETS)*100;
		//System.out.println("Percent Missing="+percentMissingData);
		setPercentMissingData(percentMissingData);
		/*if (this.scanCounter < this.scansPerGranule) {
			if (stats == null) {
				System.out.println("Deleting partial ATMS Granule -- [" + this.scanCounter + "] scans, collected packets [" + this.sciPacketCount + "]");
			} else {
				stats.sci_discardedGranules.value++;
			}	
			ret = false;
		} else*/ {
			if (stats == null) {
				//System.out.println("Creating ATMS granule of [" + this.scanCounter + "] scans");
			} else {
				stats.sci_createdGranules.value++;
			}
			super.write(hdfFile);
			ret = true;
			
		}
		
		//System.out.println("ATMS Sci packets in granule: " + sciPacketCount + " Packet per scan: " + (float)sciPacketCount / (float)scansPerGranule);
		//System.out.println("ATMS app ids = " + super.appIdCountersToString(scansPerGranule));
		scanCount = -1;
		
		sciPacketCount = 0;

		return ret;
	}
	
	/**
	 * Close out the RawApplicationPacket which writes the results to the HDF file and cleans up.
	 */
	public void close() throws RtStpsException {
		super.close();
		scanCount = -1;
		sciPacketCount = 0;
	}

}
